@@ -325,13 +325,13 @@
 (defun list-all-digests ()
   (loop for symbol being each external-symbol of (find-package :ironclad)
      if (digestp symbol)
-     collect symbol into digests
+     collect (intern (symbol-name symbol) :keyword) into digests
      finally (return (sort digests #'string<))))
 
 (defun digest-supported-p (name)
   "Return T if the digest NAME is a valid digest name."
   (and (symbolp name)
-       (not (null (digestp name)))))
+       (not (null (digestp (massage-symbol name))))))
 
 (defmethod digest-length ((digest-name symbol))
   (or (digestp (massage-symbol digest-name))
